
package org.gnu.composer.CommonTools;


import java.io.File;
import java.io.FileFilter;


/**
 * Purpose: This class uses an instance of FileOnlyFilter to filter out
 * all the unwanted files in the directory. When the filter is
 * constructed, it will accept an array of strings containing all
 * the file extensions to pass through the filter. 
 * <br><br>
 * Note: the file extension arguments are not case sensitive.
 */
public class CustomFileFilter
{
	
	//the instance of the customFilter to use.
	FileOnlyFilter filter;
	
	//the directory that will be searched.
	String pathName;
	
	/**
	 * Constructor for CustomFileFilter
	 * @param path - A string describing the pathname the filter will 
	 * search under.
	 * @param extensions - an array of strings that will serve as the filter.
     *               The strings should be of file extensions to allow to pass 
     *               through the filter. Example: "txt". Do not include a '.'. 
     *               Only the extension.
	 */
	public CustomFileFilter(String path, String [] extensions)
	{
	   filter = new FileOnlyFilter(extensions);	
	   pathName = path;
	}
	
	/**
	 * Constructor for CustomFileFilter
	 * 
	 * @param path - A string describing the pathname the filter will 
	 * search under.
	 */
	public CustomFileFilter(String path)
		{
		   filter = new FileOnlyFilter();	
		   pathName = path;
		}
	
	
	/**
	 * Default Constructor for CustomFileFilter
	 */
	public CustomFileFilter()
	{
		filter = new FileOnlyFilter();
		pathName = null;
	}
	
	
	 /**
	 * Reset the extensions to pass through filter.
	 * 
	 * @param extensions - Accepts an array of strings that will serve as the filter.
     *    The strings should be of file extensions to allow to pass through
     *    the filter. Example: "txt". Do not include a '.'. Only the extension.
	 */
	public void setExtensions(String [] extensions)
	{
		filter.setExtensions(extensions);
	}
	
	

	/**
	 * Sets the path to get the files to filter from.
	 * 
	 * @param path - A string describing the pathname the filter will
     * search under.
	 */
	public void setPathName(String path)
	{
		pathName = path;
	}
	
	
	/**
	 * Filters the files in the given directory, returning
	 * an array of the files that passed the filter.
	 * <br><br>
	 * This method will be what looks into the directory, filter
     * files and subdirectories out, and return File objects for every
     * file that passes the current filter's settings. use setPathName 
     * and setExtensions to change these settings.
	 * 
	 * @return An array of File objects that passed the current filter
     * settings. If something goes wrong (directory to search is invalid, 
     * or the SecurityException is thrown), null will be returned.
	 */
	public File [] getFiles()
	{
		try{	
			File activeFilter = new File(pathName);
			return activeFilter.listFiles(filter);
		}
		
		//could not get IO access to the directory described.
		catch(SecurityException exception)
		{
			//LOG.info("SecurityException while trying to list files in a directory using the following filter: " + filter.toString(), exception);
			return null;
		}

	}
}

class FileOnlyFilter implements FileFilter
{
//	all of the extension to pass. If null, passes all extensions.
	//extensions are case sensitive.
	String [] validExtensions;
	
	
	/**
	 * Constructor for the filter
	 * <br><br>
	 * Accepts an array of strings. Each element is an extension
     * to accept. Example extension: "txt". Do not include the '.'.
	 * 
	 * @param extensions - extensions allowed to pass the filter
	 */
	FileOnlyFilter(String [] extensions)
	{
		validExtensions = extensions;
	}
		
	/**
	 * CustomFilter: Default constructor for the customFilter class
	 */
	FileOnlyFilter()
	{
		validExtensions = null;
	}
		
		
	/**
	 * Will set what extensions to pass through the filter.
	 * 
	 * @param extensions - all extensions to allow filter to pass
	 */
	public void setExtensions(String [] extensions){
		validExtensions = extensions;
	}
		
	
	/**
	 * getFileExtension: Will extract file extension from File object.
	 * 
	 * @param inFile - File to extract the file extension from
	 * 
	 * @return String - file's extension. null if the file has none.
	 */
	public static String getFileExtension(File inFile) {
	   String extension = null;
	   if(inFile != null)
	   {
		   String fileName = inFile.getName();
		   int i = fileName.lastIndexOf('.');
		   if (i > 0 &&  i < fileName.length() - 1) 
		   {
			   extension = fileName.substring(i+1);
		   }
	   }
	   return extension;
	}
	
	/**
	 * Decides if a file passes the filter or not.
	 * <br><br>
	 * Required to be defined by the FileFilter class, this method
	 * _is_ the filter. It is an implimentation of what the filter should
	 * act like.
	 * 
	 * @param inFile - File object to be tested.
	 * 
	 * @return boolean - if file passed filter or not.
	 */
	public boolean accept(File inFile)
	{

	   //only return files.
	   if (inFile.isFile()) 
	   {
	   		//if validExtensions is null, we pass the file
	   		if (validExtensions != null)
	   		{
	   			String fileExtension = getFileExtension(inFile);
	   		
	   			if(fileExtension != null)
	   			{
	   				//go through all the extensions until either
	   				//I run out or one matches.
	   				
	   				for(int i=0;i<validExtensions.length;i++)
	   				{
	   					if(validExtensions[i].toLowerCase().equals(fileExtension.toLowerCase()))
	   					{
	   						return true;
	   					}
	   				}
	   			
	   				//if it is not in the list, it is not passed
	   				return false;
	   			}
	   		
	   			//the filter is on, and only files with file
	   			//extensions pass. This does not have a file 
	   			//extension.
	   			else
	   			{
	   				return false;
	   			}
	   		}
	   	
	   		//validExtensions is null, so pass.
	   		else
	   		{
	   			return true;
	   		}
	   }
	   
	   //file is a directory, so it does not pass the filter.
	   else
	   {
	   		return false;
	   }  
	}

	
	public String toString()
	{
		String toReturn;;
		
		if(validExtensions == null)
		{
			toReturn = "null filter";
		}
		else
		{
			toReturn = "{";
			
			for(int index = 0; index < validExtensions.length; index++)
			{
				toReturn = toReturn + " " + validExtensions[index];
			}
			
			toReturn = toReturn + " }";

		}
		
		return toReturn;
	}
	
	
	
	/**
	 * getDescription: Returns description of the filter.
	 * <br><br>
	 * Required by the FileFilter class. It will never be used
     * by this class.
	 * 
	 * @return String - Description of Filter class
	 */
	public String getDescription() {
	   return "Custom Filtered Files";
	}
}