

package org.gnu.composer.CommonTools;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;

import org.gnu.composer.RuntimeSettings.*;

/**
* This class provides a logging mechanism for BravoProController.
* The log must be started, and can then be added to.  
* 
* @author Branden Archer
*/
public class Logger
{
	/**
	 * The directory the log file can be found in.
	 */
	public static final String DEFAULT_LOG_DIRETORY = RuntimeSettings.logRoot;
	private static String logDirectory_ = DEFAULT_LOG_DIRETORY;
	
	private static File currentLogFile_ = null;
	/**
	 * The FileWriter variable that writes the log.
	 */
	private static FileWriter currentLogFileWriter_ = null;
	private static boolean logOpen_ = false;
	private static Object writtingMutex_ = new Object();
	
	private static String baseFilename_ = null;
	
	//The number of log files currently (ie, the current log file used)
	private static long numOfLogFiles_ = 1;
	
	//The maximum desired size of a log file, in bytes
	private final static long ONE_MEGABYTE = 1024 * 1024;
	private static long maxLogFileSize_ = ONE_MEGABYTE;
	
	/*
	 * There will be certain levels of logs desired after startup.
	 * Each option can be selected or de-selected with 
	 * setLoggingOptions
	 */
	private static boolean logDebug_    = true;
	private static boolean logWarning_  = true;
	private static boolean logError_    = true;
	private static boolean logFirmware_ = true;
	private static boolean logEngine_   = true;
	private static boolean logDb_       = true;
	private static boolean logRipper_   = true;
	private static boolean logGui_      = true;
	private static boolean logSystem_   = true;
	
	/*
	 * Should the logger auto format the logs, putting links between
	 * the logs?
	 */
	private static boolean autoFormatLogs_ = false;
	private static String lastLogFile_ = null;
	private static String nextLogFile_     = null;
	
	/**
	 * Set the directory where the log files will be put.
	 * If the field "newDirectory" is not a valid 
	 * directory, the setting is ignored.
	 * @param newDirectory
	 * @return true iff the setting was successfully made
	 */
	public static boolean setLogDirectory(String newDirectory)
	{
		boolean success = false;
		
		File directory = new File(newDirectory);
		if(directory.exists() && directory.isDirectory())
		{
			logDirectory_ = newDirectory;
			success = true;
		}
		
		return success;
	}
	
	/**
	 * Sets the max log size before the logs are split.
	 * @param maxLogFileSizeInBytes
	 * @return true iff the settings were valid and successfully
	 * set.
	 */
	public static boolean setMaxLogFileSize(long maxLogFileSizeInBytes)
	{
		boolean success = false;
		
		if(maxLogFileSizeInBytes > 0)
		{
			maxLogFileSize_ = maxLogFileSizeInBytes;
			success = true;
		}
		
		return success;
	}
	
	/**
	 * Set if the logs should be autoformatted or not. This can only
	 * be set when the logger is off.
	 * @param autoFormatLogs
	 */
	public static void autoFormatLogs(boolean autoFormatLogs)
	{
		if(logOpen_ == false)
		{
			synchronized(writtingMutex_)
			{
				autoFormatLogs_ = autoFormatLogs;
			}
		}
	}
	
	/**
	 * Starts the logger.
	 *   -creates the log directory if it does not exist.
	 *   -opens the log file for writing.
	 *   -date and time stamps the opening of the log.
	 *   -assumes default logging (everything)
	 */
	public static void startLogger()
	{
		if(logOpen_ == false)
		{
			synchronized(writtingMutex_)
			{
				try
				{
					String date = getCurrentTimeAsString();
					date = date.replace(":",".");
					date = date.replace("/",".");
					date = date.replace(" ","_");
					File logDir = new File(logDirectory_);

					if(logDir.exists() == false)
					{
						try
						{
							logDir.mkdirs( );
						}
						catch(SecurityException e)
						{
							e.printStackTrace();
						}
					}
					
					currentLogFile_ = new File(logDirectory_+"log_"+date+".log");
					baseFilename_   = "log_"+date+"_(1).log";
					lastLogFile_    = logDirectory_+"log_"+date+".log";
					nextLogFile_    = logDirectory_+"log_"+date+"_(2).log";
					currentLogFile_.createNewFile();
					currentLogFileWriter_ = new FileWriter(currentLogFile_);

					commitToLog("Log started "+getCurrentTimeAsString()+"\r\n");
					
					logOpen_ = true;
				}
				catch(IOException e)
				{
					e.printStackTrace();
					logOpen_ = false;
				}
			}
		}
	}
	
	
	/**
	 * Starts the logger.
	 *   -creates the log directory if it does not exist.
	 *   -opens the log file for writing.
	 *   -date and time stamps the opening of the log.
	 *   -assumes default logging (everything)
	 *  <br>
	 *  The log file will be put into the default log directory and will
	 *  be appended with ".log" at the end.
	 *  @param fileNamePrefix - what the log file should start with. ".log" will
	 *  be appended to the log file name.
	 *   @param includeData - True means to append the date/time to the
	 *   end of the passed in name before ".log".
	 */
	public static void startLogger(String fileNamePrefix, boolean includeDate)
	{
		if(logOpen_ == false)
		{
			synchronized(writtingMutex_)
			{
				try
				{
					String finalLogFileName = null;
					
					File logDir = new File(logDirectory_);
					if(logDir.exists() == false)
					{
						try
						{
							logDir.mkdirs( );
						}
						catch(SecurityException e)
						{
							
						}
					}
					
					if(includeDate)
					{
						String date = null;
						date = getCurrentTimeAsString();
						date = date.replace(":",".");
						date = date.replace("/",".");
						date = date.replace(" ","_");
						finalLogFileName = fileNamePrefix + "_" + date + ".log";
					}
					else
					{
						finalLogFileName = fileNamePrefix + ".log";
					}
					
					currentLogFile_ = new File(logDirectory_ + finalLogFileName);
					baseFilename_ = fileNamePrefix + "_(1).log";
					currentLogFile_.createNewFile();
					currentLogFileWriter_ = new FileWriter(currentLogFile_);
					commitToLog("Log started "+getCurrentTimeAsString()+"\r\n");
					
					logOpen_ = true;
				}
				catch(IOException e)
				{
					logOpen_ = false;
				}
			}
		}
	}

	/**
	 * Checks the size of the current log file. If it is over
	 * a certain threshold, the log file is stopped and
	 * a new file is opened.
	 */
	private static void checkLogSize()
	{
		String lastLogFileName;
		String nextLogFileName;
		
		if(currentLogFile_.length() >= maxLogFileSize_)
		{
			try
			{
				String dateAndTime = getCurrentTimeAsString();
				writeToLog(dateAndTime+"  Log File Split\r\n");
				
				if(autoFormatLogs_)
				{
					nextLogFileName = 
						nextLogFile_.replace(".log",".html"); 
					
				}
				else
				{
					nextLogFileName = nextLogFile_;
				}
				
				writeToLog("NextLog: " + nextLogFileName + "\r\n");

				currentLogFileWriter_.close();
				
				lastLogFileName = lastLogFile_;
		
				baseFilename_ = baseFilename_.replace(
						"(" + numOfLogFiles_ + ")", 
						"(" + (numOfLogFiles_+1) + ")");
				
				numOfLogFiles_++;
				
				currentLogFile_ = new File(logDirectory_+baseFilename_);
				currentLogFile_.createNewFile();
				currentLogFileWriter_ = new FileWriter(currentLogFile_);
				
				writeToLog("PrevLog: " + lastLogFileName + "\r\n");
				
				lastLogFile_ = nextLogFile_;
				String next = baseFilename_ = baseFilename_.replace(
						"(" + numOfLogFiles_ + ")", 
						"(" + (numOfLogFiles_+1) + ")");
				nextLogFile_ = logDirectory_ + next;
				
				writeToLog(dateAndTime+"  Log File Continued\r\n");
			}
			catch(IOException e)
			{
				
			}
		
		}
	}

	/**
	 * Sets what should be logged.
	 */
	public static void setLoggingOptions(boolean warnings, boolean errors, boolean debug,
			boolean firmware, boolean engine, boolean db, boolean ripper, boolean gui,
			boolean system)
	{
		logWarning_  = warnings;
		logError_    = errors;
		logDebug_    = debug;
		logFirmware_ = firmware;
		logEngine_   = engine;
		logDb_       = db;
		logRipper_   = ripper;
		logGui_      = gui;
		logSystem_   = system;
		
		if(logOpen_)
		{
			String toWrite = "";
			
			if(currentLogFileWriter_ != null)
			{
				String dateAndTime = getCurrentTimeAsString();
				
				toWrite += dateAndTime+"  Log Level set to: Crash ";
				
				if(logError_)
				{
					toWrite += "Errors ";
				}
				if(logWarning_)
				{
					toWrite += "Warnings ";
				}
				if(logDebug_)
				{
					toWrite += "Debug ";
				}
				if(logFirmware_)
				{
					toWrite += "Firmware ";
				}
				if(logEngine_)
				{
					toWrite += "Engine ";
				}
				if(logDb_)
				{
					toWrite += "DB ";
				}
				if(logRipper_)
				{
					toWrite += "Ripper ";
				}
				if(logGui_)
				{
					toWrite += "Gui ";
				}
				if(logSystem_)
				{
					toWrite += "System ";
				}
				
				toWrite += "\r\n";

				commitToLog(toWrite);
			}
		}
	}
	
	/**
	 * Closes the logger.
	 * 
	 * It is ok if this method is not invoked - each line is flushed
	 * as it is written, so the log file will be accessible even if 
	 * the File.close() method is not called on it (which is all this
	 * method does).
	 */
	public static void closeLogger()
	{
		if(logOpen_)
		{
			synchronized(writtingMutex_)
			{
				try
				{
					currentLogFileWriter_.close();
					currentLogFileWriter_ = null;
				}
				catch(IOException e)
				{
					
				}
			}
			
			logOpen_ = false;
		}
	}
	
	/**
	 * Adds a time-stamped entry to the log. 
	 * Message is an error.
	 * @param logType
	 * 			  Type of log message to add.
	 * @param sourceClass
	 *            String object reference to the class calling
	 *            the addToLog method (can be any string)
	 * @param sourceMethod
	 *            String object reference to the job this entry is
	 *            associated with (can be any string)
	 * @param message
	 *            String object of actual log message
	 *            (can be any string)
	 */
	public static void addLog(LogTypes logType, String sourceClass, 
			String sourceMethod, String message)
	{
		if(logOpen_)
		{
			String dateAndTime = getCurrentTimeAsString();
			final String SEPERATOR = ":  ";
			String sourceInfo = sourceClass + SEPERATOR 
				+ sourceMethod + SEPERATOR;
			
			final String UNINITIALIED 
				= "BUG: A blank message was written from the Logger";
			
			String toWrite = UNINITIALIED;
			boolean writeMessage = false;
			
			switch(logType.value())
			{
				//When code debug information is wanted
				case LogTypes.DEBUG_VALUE:
					if(logDebug_)
					{
						writeMessage = true;
					}
					break;
					
				//Strange events have occured, but they are recoverable
				case LogTypes.WARNING_VALUE:
					if(logWarning_)
					{
						writeMessage = true;
					}
					break;
					
				//Problems have occured and bad things are resulting
				case LogTypes.ERROR_VALUE:
					if(logError_)
					{
						writeMessage = true;
					}
					break;
					
				//Components of the system have failed and quit working
				case LogTypes.CRASH_VALUE:
					writeMessage = true;
					break;
					
				//A condition has occured which only means that 
				//a bug has been found.
				case LogTypes.BUG_VALUE:
					writeMessage = true;
					break;
					
				default:
					writeMessage = true;
					break;
			}
			
			if(writeMessage)
			{
				toWrite = dateAndTime + "  " + logType + ": " + sourceInfo + message + "\r\n";
				commitToLog(toWrite);
			}
		}
	}
	
	
	/**
	 * Prints out an exception to the log. 
	 * This entry is assumed to be a horrible
	 * system crash.
	 * @param programName - the name of the program which has crashed
	 * @param exceptionToPrint
	 */
	public static void addCrashToLog(String programName, Exception exceptionToPrint)
	{
		if(logOpen_)
		{
			String toWrite = "";
			
			String dateAndTime = getCurrentTimeAsString();

			toWrite += "\r\n" + dateAndTime + ": " 
				+ programName + " has crashed!\r\n";
			
			String exceptionClass = exceptionToPrint.getClass().getName();
			if(exceptionClass != null)
			{
				toWrite += "     " + exceptionClass + "\r\n";
			}
			
			StackTraceElement [] trace = exceptionToPrint.getStackTrace();
			
			for(int line = 0; line < trace.length; line++)
			{
				toWrite += "     at " + trace[line].toString() + "\r\n";
			}
			
			toWrite += "\r\n";
			
			commitToLog(toWrite);
		}
	}

	/**
	 * Prints out an exception to the log. 
	 * This entry is assumed to be not life
	 * threatening, but bad enough to look into.
	 * 
	 * @param exceptionToPrint
	 * @param message - explination of what was being done at the
	 * time of the exception.
	 */
	public static void addExceptionToLog(Exception exceptionToPrint, String message)
	{
		if(logOpen_)
		{
			String toWrite = "";
			
			String dateAndTime = getCurrentTimeAsString();

			toWrite += "\r\n" + dateAndTime 
					+ ": Caught Exception has occured.\r\n"
					+ message + "\r\n";
			
			String exceptionClass = exceptionToPrint.getClass().getName();
			if(exceptionClass != null)
			{
				toWrite += "     " + exceptionClass + "\r\n";
			}
			
			StackTraceElement [] trace = exceptionToPrint.getStackTrace();
			
			for(int line = 0; line < trace.length; line++)
			{
				toWrite += "     at " + trace[line].toString() + "\r\n";
			}
			
			toWrite += "\r\n";
			
			commitToLog(toWrite);
		}
	}
	
	/**
	 * Writes the desired message to the log after first
	 * checking the log's current size.
	 * @param toWrite - message to write to log.
	 */
	private static void commitToLog(String toWrite)
	{
		synchronized(writtingMutex_)
		{
			checkLogSize();
			writeToLog(toWrite);
		}
	}
	
	/**
	 * Write a message directly to the log without first checking
	 * the size.
	 * @param toWrite - message to write to log.
	 */
	private static void writeToLog(String toWrite)
	{
		synchronized(writtingMutex_)
		{
			try
			{
				currentLogFileWriter_.append(toWrite);
				currentLogFileWriter_.flush();
			}
			catch(IOException e)
			{
				e.printStackTrace();
			}
		}
	}
	
	/**
	 * Retrieve the current file to which logs are being written.
	 * @return log file
	 */
	public static File currentLogFile()
	{
		//This is synchronized, as if the logs are being written to, and
		//the logs are actually being split, an incorrect value of the
		//'current log file' could be returned. This ensures that any
		//split action can occur first.
		synchronized(writtingMutex_)
		{
			return currentLogFile_;	
		}
	}
	
	/**
	 * Returns the given 24-hour time as a formatted string
	 * <br><br>
	 * Time output format:<br>
	 * MM/DD/YYYY hh:mm:ss<br>
	 * 
	 * @param time - time to convert to a string. Use
	 * getCurrentTimeAsString() to get the current time.
	 * @return String - The formatted time.
	 */
	public static String makeFormattedTimeString(long time)
	{
		
		//Time output format:
		//MM/DD/YYYY hh:mm:ss
		
		int MM,DD,YYYY,hh,mm,ss;
		String sMM,sDD,sYYYY,shh,smm,sss;
		
		Date myDate = new Date();
		myDate.setTime(time);
		Calendar myCalendar = new GregorianCalendar();
		myCalendar.setTime(myDate);
		
		MM = myCalendar.get(Calendar.MONTH) + 1; //Jan=0, Feb=1, ...
		DD = myCalendar.get(Calendar.DAY_OF_MONTH);
		YYYY = myCalendar.get(Calendar.YEAR);
		hh = myCalendar.get(Calendar.HOUR_OF_DAY);
		mm = myCalendar.get(Calendar.MINUTE);
		ss = myCalendar.get(Calendar.SECOND);
		
		//convert month to a string
		if(MM <= 9)
		{
			sMM= "0" + Integer.toString(MM,10);
		}
		else
		{
			sMM= Integer.toString(MM,10);
		}
		
		//convert day to a string
		if(DD <= 9)
		{
			sDD = "0" + Integer.toString(DD,10);
		}
		else
		{
			sDD = Integer.toString(DD,10);
		}
		
		//convert year to a string
		sYYYY = Integer.toString(YYYY,10);
		
		
		//convert hours to a string
		if(hh <= 9)
		{
			shh = "0" + Integer.toString(hh,10);
		}
		else
		{
			shh = Integer.toString(hh,10);
		}
		
		//convert minutes to a string
		if(mm <= 9)
		{
			smm = "0" + Integer.toString(mm,10);
		}
		else
		{
			smm = Integer.toString(mm,10);
		}
		
		//convert seconds to a string
		if (ss <= 9)
		{
			sss = "0" + Integer.toString(ss,10);
		}
		else
		{
			sss = Integer.toString(ss,10);
		}
		
        //Time output format:
		//MM/DD/YYYY hh:mm:ss
		
		return sMM + "/" + sDD + "/" + sYYYY + " " + shh + ":" 
		     + smm + ":" + sss;
		
	}
	
	/**
	 * @return formatted string containing the current time
	 */
	public static String getCurrentTimeAsString()
	{
		return makeFormattedTimeString(System.currentTimeMillis());
	}
	
}
