package org.gnu.composer.DB;

import java.sql.*;
import java.util.Properties;

/**
 * This class contains methods for accessing the
 * db. All methods are synchronized such that only
 * one thread can modify the db at any time. All
 * db accessing should be done in this class, with
 * user friendly front-ends else where in the code.
 * 
 * @author Branden Archer
 */
public class DBAccess 
{
	private static Connection conn;
	
	/**
	 * Load up the driver class for the db
	 */
	static
	{
		try
		{
			Class.forName("org.postgresql.Driver");
			
			String url = "jdbc:postgresql://localhost/composer";
			Properties props = new Properties();
			props.setProperty("user","quatre");
			props.setProperty("password","");
			conn = DriverManager.getConnection(url, props);
			conn.setAutoCommit(true);
		}
		catch(Exception e)
		{
			e.printStackTrace();
			System.exit(1);
		}
	}
	
	/**
	 * Queries the db for the next jobid to use. The next
	 * id is retrieved, and then the value in the db is
	 * incrimented.
	 * @return the next jobid. -1 if an error occured
	 */
	synchronized public static int nextJobid()
	{
		int nextJobid = -1;
		
		try
		{
			Statement statement = conn.createStatement();
			ResultSet results = statement.executeQuery(
				"SELECT " + DBConsts.CONFIG_VALUE + " FROM " + DBConsts.CONFIG_TABLE
				+ " where " + DBConsts.CONFIG_NAME + " = '" + DBConsts.CONFIG_NEXT_JOBID + "';");
			if (results.next()) 
			{
				nextJobid = Integer.parseInt(results.getString(1));
			}
			results.close();
			statement.close();
			
			setNextJobid(nextJobid + 1);
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
		
		return nextJobid;
	}
	
	/**
	 * Sets the next jobid in the db.
	 * @param newValue - the value for the next jobid
	 */
	synchronized public static void setNextJobid(int newValue)
	{
		try
		{
			Statement statement = conn.createStatement();
			statement.execute(
				"update " + DBConsts.CONFIG_TABLE + 
				" set "   + DBConsts.CONFIG_VALUE + "="  + newValue + 
				" where " + DBConsts.CONFIG_NAME  + "='" + DBConsts.CONFIG_NEXT_JOBID + "';");
			statement.close();
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
	}
	
	/**
	 * Queries the db for the next backup to use. The
	 * current value is retrieved, then the value in the
	 * db is incremented.
	 * @return the next jobid. -1 if an error occured
	 */
	synchronized public static int nextBackupid()
	{
		int nextBackupid = -1;
		
		try
		{
			Statement statement = conn.createStatement();
			ResultSet results = statement.executeQuery(
				"SELECT " + DBConsts.CONFIG_VALUE + " FROM " + DBConsts.CONFIG_TABLE
				+ " where " + DBConsts.CONFIG_NAME + " = '" + DBConsts.CONFIG_NEXT_BACKUPID + "';");
			if (results.next()) 
			{
				nextBackupid = Integer.parseInt(results.getString(1));
			}
			results.close();
			statement.close();
			
			setNextBackupid(nextBackupid + 1);
		
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
		
		return nextBackupid;
	}
	
	/**
	 * Sets the next backup in the db.
	 * @param newValue - the value for the next jobid
	 */
	synchronized public static void setNextBackupid(int newValue)
	{
		try
		{
			Statement statement = conn.createStatement();
			statement.execute(
				"update " + DBConsts.CONFIG_TABLE + 
				" set "   + DBConsts.CONFIG_VALUE + "="  + newValue + 
				" where " + DBConsts.CONFIG_NAME  + "='" + DBConsts.CONFIG_NEXT_BACKUPID + "';");
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
	}
	
	/**
	 * Add a job if it is in the db, otherwise update the
	 * job information in the db.
	 * @param job - job to add/update
	 * @return True iff all went well
	 */
	synchronized public static boolean addOrUpdateJob(Job job)
	{
		boolean success = true;
		
		try
		{
			Statement statement = conn.createStatement();
			String sql;
			int updatedRows;
			
			//First update the instance in the activejobs lists
			sql = 
				"update " + DBConsts.ACTIVE_JOB_TABLE + 
				" set " + 
				DBConsts.ACTIVE_JOB_STATE       + "="  + job.state().stateValue() + ","  + 
				DBConsts.ACTIVE_JOB_LABEL       + "='" + job.label()              + "'," + 
				DBConsts.ACTIVE_JOB_TOTAL_DISCS + "="  + job.totalDiscs()         + ","  + 
				DBConsts.ACTIVE_JOB_DONE_DISCS  + "="  + job.doneDiscs()          + ","  + 
				DBConsts.ACTIVE_JOB_ERROR_DISCS + "="  + job.errorDiscs()         + " "  + 
				"where " + 
				DBConsts.ACTIVE_JOB_JOBID       + "="  + job.jobid()              + ";";

			updatedRows = statement.executeUpdate(sql);
			
			if(updatedRows == 0)
			{
				//Add the job
				sql = 
					"insert into " + DBConsts.ACTIVE_JOB_TABLE + 
					"("                                        + 
					DBConsts.ACTIVE_JOB_JOBID       + ","      +  
					DBConsts.ACTIVE_JOB_STATE       + ","      + 
					DBConsts.ACTIVE_JOB_TOTAL_DISCS + ","      +
					DBConsts.ACTIVE_JOB_DONE_DISCS  + ","      + 
					DBConsts.ACTIVE_JOB_ERROR_DISCS + ","      + 
					DBConsts.ACTIVE_JOB_LABEL                  + 
					") values ("                               + 
					job.jobid()                     + ","      + 
					job.state().stateValue()        + ","      + 
					job.totalDiscs()                + ","      +
					job.doneDiscs()                 + ","      + 
					job.errorDiscs()                + ","      +
					"'" + job.label() + "'"                    + 
					");";

				statement.execute(sql);
			}
			
			//Drop the individual job table so the latest backup items can be added
			try
			{
				statement.execute(
					"delete from " + DBConsts.INDIVIDUAL_JOB_TABLE + job.jobid() + ";");
			}
			catch(SQLException e)
			{
				//The table did not exist. Add it now.
				sql = 
					"create table " + DBConsts.INDIVIDUAL_JOB_TABLE + job.jobid() 
					+ "(" + 
					DBConsts.INDIVIDUAL_JOB_BACKUP + " int" + 
					");";

				statement.execute(sql);
			}
			
			//Add all the backup id into the specific job table
			Integer [] allBackupIds = job.backups();
			for(int index = 0; index < allBackupIds.length; index++)
			{
				sql = 
					"insert into " + DBConsts.INDIVIDUAL_JOB_TABLE + job.jobid() + 
					"("                            + 
					DBConsts.INDIVIDUAL_JOB_BACKUP + 
					") values ("                   + 
					allBackupIds[index].intValue() + 
					");";
				statement.execute(sql);
			}
			
			statement.close();
		}
		catch(SQLException e)
		{
			e.printStackTrace();
			success = false;
		}
		
		return success;
	}
	
	/**
	 * Retrieve a job from the db if it exists
	 * @param jobid - jobid of the job to retrieve
	 * @return the job, or null if the job dne.
	 */
	synchronized public static Job lookupJob(int jobid)
	{
		Job job = null;
		String sql;
		Statement statement;
		ResultSet results;
		
		final int STATE_COLUMN       = 1;
		final int TOTAL_DISCS_COLUMN = 2;
		final int DONE_DISCS_COLUMN  = 3;
		final int ERROR_DISCS_COLUMN = 4;
		final int LABEL_COLUMN       = 5;
		
		try
		{
			statement = conn.createStatement();
			//Get the general job information
			sql = 
				"select "                             + 
				DBConsts.ACTIVE_JOB_STATE       + "," + //Column 1
				DBConsts.ACTIVE_JOB_TOTAL_DISCS + "," + //Column 2
				DBConsts.ACTIVE_JOB_DONE_DISCS  + "," + //Column 3
				DBConsts.ACTIVE_JOB_ERROR_DISCS + "," + //Column 4
				DBConsts.ACTIVE_JOB_LABEL             + //Column 5
				" from " + DBConsts.ACTIVE_JOB_TABLE  +
				" where " + DBConsts.ACTIVE_JOB_JOBID + "=" + jobid + ";";
			
			results = statement.executeQuery(sql);
			
			//If the job was found, make a new job and load in the values
			if(results.next())
			{
				job = new Job(jobid);
				job.state(
						JobState.convFromInt(
								results.getInt(STATE_COLUMN)));
				job.totalDiscs(
						results.getInt(TOTAL_DISCS_COLUMN));
				job.doneDiscs(
						results.getInt(DONE_DISCS_COLUMN));
				job.errorDiscs(
						results.getInt(ERROR_DISCS_COLUMN));
				job.label(
						validate(
								results.getString(LABEL_COLUMN)));
			}
			
			results.close();
			
			//If the job was found, now load in the specific backup instances.
			if(job != null)
			{
				final int ID_COLUMN = 1;
				sql = "select " + DBConsts.INDIVIDUAL_JOB_BACKUP 
					+ " from " + DBConsts.INDIVIDUAL_JOB_TABLE + job.jobid() + ";";
				results = statement.executeQuery(sql);
				while(results.next())
				{
					int id = results.getInt(ID_COLUMN);
					job.addBackupId(id);
				}
			}
			
			statement.close();
		}
		catch(SQLException e)
		{
			
		}
		
		return job;
	}
	
	/**
	 * Deletes a job from the database if it exists.
	 * @param jobid - jobid of the job to delete.
	 * @return true iff the job existed and was 
	 * successfully deleted. False otherwise.
	 */
	synchronized public static boolean deleteJob(int jobid)
	{
		boolean success = true;
		
		try
		{
			Statement statement = conn.createStatement();
			String sql = 
				"delete from " + DBConsts.ACTIVE_JOB_TABLE + 
				" where " + DBConsts.ACTIVE_JOB_JOBID + " = " + jobid + ";";
			if(statement.executeUpdate(sql) != 1)
			{
				System.out.println("Nothing deleted");
				success = false;
			}
			
			try
			{
				sql = "drop table " + DBConsts.INDIVIDUAL_JOB_TABLE + jobid + ";";
				statement.execute(sql);
			}
			catch(SQLException e)
			{
				//The table was not there to delete.
				e.printStackTrace();
			}
			
			statement.close();
		}
		catch(SQLException e)
		{
			e.printStackTrace();
			success = false;
		}
		
		return success;
	}
	
	/**
	 * Add a backup if it is in the db, otherwise update the
	 * backup information in the db.
	 * @param backup - backup to add/update
	 * @return True iff all went well
	 */
	synchronized public static boolean addOrUpdateBackup(Backup backup)
	{
		boolean success = true;
		
		try
		{
			Statement statement = conn.createStatement();
			String sql;
			int updatedRows;
			
			//First update the instance in the activejobs lists
			sql = 
				"update " + DBConsts.BACKUP_TABLE + 
				" set "   + 
				DBConsts.BACKUP_NAME          + "=" + "'" + backup.name()         + "'" + "," + 
				DBConsts.BACKUP_DISC_SIZE     + "=" +     + backup.discsize()           + "," + 
				DBConsts.BACKUP_NEW           + "=" +       backup.isnew()              + "," + 
				DBConsts.BACKUP_CREATION_DATE + "=" + "'" + backup.creationdate() + "'" + " " + 
				"where " + 
				DBConsts.BACKUP_BACKUP_ID + "="  + backup.backupid() + ";";

			updatedRows = statement.executeUpdate(sql);
			
			if(updatedRows == 0)
			{
				//Add the job
				sql = 
					"insert into " + DBConsts.BACKUP_TABLE  + 
					"("                                     + 
					DBConsts.BACKUP_BACKUP_ID         + "," +  
					DBConsts.BACKUP_NAME              + "," + 
					DBConsts.BACKUP_NEW               + "," +
					DBConsts.BACKUP_DISC_SIZE         + "," +
					DBConsts.BACKUP_CREATION_DATE     + " " + 
					") values ("                            + 
					      backup.backupid()           + "," + 
					"'" + backup.name()         + "'" + "," + 
					      backup.isnew()              + "," + 
					    + backup.discsize()           + "," +
					"'" + backup.creationdate() + "'" + " " +
					");";

				statement.execute(sql);
			}
			
			//Drop the individual disc table so the latest track items can be added
			try
			{
				statement.execute(
					"delete from " + DBConsts.INDIVIDUAL_DISC_TABLE + backup.backupid() + ";");
			}
			catch(SQLException e)
			{
				//The table did not exist. Add it now.
				sql = 
					"create table " + DBConsts.INDIVIDUAL_DISC_TABLE + backup.backupid() 
					+ "(" + 
					DBConsts.INDIVIDUAL_DISC_ARTIST     + " varchar(256), " +
					DBConsts.INDIVIDUAL_DISC_ALBUM      + " varchar(256), " +
					DBConsts.INDIVIDUAL_DISC_TRACKNAME  + " varchar(256), " +
					DBConsts.INDIVIDUAL_DISC_TRACKNUM   + " int, "          +
					DBConsts.INDIVIDUAL_DISC_DURATION   + " varchar(256), " +
					DBConsts.INDIVIDUAL_DISC_FORMAT     + " varchar(256), " +
					DBConsts.INDIVIDUAL_DISC_BITRATE    + " int, "          +
					DBConsts.INDIVIDUAL_DISC_GENRE      + " varchar(256), " +
					DBConsts.INDIVIDUAL_DISC_YEAR       + " int, "          +
					DBConsts.INDIVIDUAL_DISC_SAMPLERATE + " int, "          +
					DBConsts.INDIVIDUAL_DISC_COMMENT    + " varchar(256)  " +
					");";

				statement.execute(sql);
			}
			
			//Add all the tracks into the specific disc table
			Track [] allTracks = backup.disc().tracks();

			for(int index = 0; index < allTracks.length; index++)
			{
				sql = 
					"insert into " + DBConsts.INDIVIDUAL_DISC_TABLE + backup.backupid() + 
					"("                                       + 
					DBConsts.INDIVIDUAL_DISC_ARTIST           + ", " + 
					DBConsts.INDIVIDUAL_DISC_ALBUM            + ", " + 
					DBConsts.INDIVIDUAL_DISC_TRACKNAME        + ", " + 
					DBConsts.INDIVIDUAL_DISC_TRACKNUM         + ", " + 
					DBConsts.INDIVIDUAL_DISC_DURATION         + ", " + 
					DBConsts.INDIVIDUAL_DISC_FORMAT           + ", " + 
					DBConsts.INDIVIDUAL_DISC_BITRATE          + ", " + 
					DBConsts.INDIVIDUAL_DISC_GENRE            + ", " + 
					DBConsts.INDIVIDUAL_DISC_YEAR             + ", " + 
					DBConsts.INDIVIDUAL_DISC_SAMPLERATE       + ", " + 
					DBConsts.INDIVIDUAL_DISC_COMMENT          + "  " + 
					") values ("                                     + 
					"'" + allTracks[index].artist()     + "'" + ", " + 
					"'" + allTracks[index].album()      + "'" + ", " + 
					"'" + allTracks[index].trackname()  + "'" + ", " + 
				 	      allTracks[index].tracknum()         + ", " + 
					"'" + allTracks[index].duration()   + "'" + ", " + 
					"'" + allTracks[index].format()     + "'" + ", " + 
					      allTracks[index].bitrate()          + ", " + 
					"'" + allTracks[index].genre()      + "'" + ", " + 
					      allTracks[index].year()             + ", " + 
					      allTracks[index].samplerate()       + ", " + 
					"'" + allTracks[index].comment()    + "'" + "  " + 
					");";

				statement.execute(sql);
			}
			
			statement.close();
		}
		catch(SQLException e)
		{
			e.printStackTrace();
			success = false;
		}
		
		return success;
	}
	
	/**
	 * Retrieve a backup from the db if it exists
	 * @param backupid - backupid of the backup to retrieve
	 * @return the backup, or null if the backup dne.
	 */
	synchronized public static Backup lookupBackup(int backupid)
	{
		Backup backup = null;
		String sql;
		Statement statement;
		ResultSet results;
		
		final int NAME_COLUMN   = 1;
		final int IS_NEW_COLUMN = 2;
		final int SIZE_COLUMN   = 3;
		final int DATE_COLUMN   = 4;
		
		try
		{
			statement = conn.createStatement();
			//Get the general job information
			sql = 
				"select "                             + 
				DBConsts.BACKUP_NAME            + "," + //Column 1
				DBConsts.BACKUP_NEW             + "," + //Column 2
				DBConsts.BACKUP_DISC_SIZE       + "," + //Column 3
				DBConsts.BACKUP_CREATION_DATE         + //Column 4
				" from " + DBConsts.BACKUP_TABLE      +
				" where " + DBConsts.BACKUP_BACKUP_ID + "=" + backupid + ";";
			
			results = statement.executeQuery(sql);
			
			//If the backup was found, make a new backup and load in the values
			if(results.next())
			{
				backup = new Backup(backupid, results.getTimestamp(DATE_COLUMN));
				backup.name(
						validate(results.getString(NAME_COLUMN)));
				backup.isnew(
						results.getBoolean(IS_NEW_COLUMN));
				backup.discsize(
						results.getInt(SIZE_COLUMN));
			}
			
			results.close();
			
			//If the backup was found, now load in the specific track instances
			if(backup != null)
			{
				Disc disc = lookupDisc(backupid);
				if(disc != null)
				{
					backup.disc(disc);
				}
				else
				{
					//Disc was not found, this backup is invalid.
					backup = null;
				}
			}
			
			statement.close();
		}
		catch(SQLException e)
		{
			
		}
		
		return backup;
	}
	
	/**
	 * Retrieve a given disc given the backupid of the backup
	 * this disc belongs to.
	 * @param backupid
	 * @return Disc object, either filled or empty
	 */
	synchronized public static Disc lookupDisc(int backupid)
	{
		Disc disc = new Disc();
		String sql;
		Statement statement;
		ResultSet results;
		
		final int TRACKNAME_COLUMN  = 1;
		final int TRACKNUM_COLUMN   = 2;
		final int ALBUM_COLUMN      = 3;
		final int ARTIST_COLUMN     = 4;
		final int GENRE_COLUMN      = 5;
		final int YEAR_COLUMN       = 6;
		final int DURATION_COLUMN   = 7;
		final int BITRATE_COLUMN    = 8;
		final int SAMPLERATE_COLUMN = 9;
		final int FORMAT_COLUMN     = 10;
		final int COMMENT_COLUMN    = 11;
		
		try
		{
			statement = conn.createStatement();
			//Get the general job information
			sql = 
				"select "                                 + 
				DBConsts.INDIVIDUAL_DISC_TRACKNAME  + "," + 
				DBConsts.INDIVIDUAL_DISC_TRACKNUM   + "," + 
				DBConsts.INDIVIDUAL_DISC_ALBUM      + "," + 
				DBConsts.INDIVIDUAL_DISC_ARTIST     + "," + 
				DBConsts.INDIVIDUAL_DISC_GENRE      + "," + 
				DBConsts.INDIVIDUAL_DISC_YEAR       + "," + 
				DBConsts.INDIVIDUAL_DISC_DURATION   + "," + 
				DBConsts.INDIVIDUAL_DISC_BITRATE    + "," + 
				DBConsts.INDIVIDUAL_DISC_SAMPLERATE + "," + 
				DBConsts.INDIVIDUAL_DISC_FORMAT     + "," + 
				DBConsts.INDIVIDUAL_DISC_COMMENT          + 
				" from " + DBConsts.INDIVIDUAL_DISC_TABLE + backupid + ";";
			
			results = statement.executeQuery(sql);
			
			//If the backup was found, make a new backup and load in the values
			while(results.next())
			{
				Track track = new Track();
				track.trackname(
						validate(results.getString(TRACKNAME_COLUMN)));
				track.tracknum(
						results.getInt(TRACKNUM_COLUMN));
				track.album(
						validate(results.getString(ALBUM_COLUMN)));
				track.artist(
						validate(results.getString(ARTIST_COLUMN)));
				track.genre(
						validate(results.getString(GENRE_COLUMN)));
				track.year(
						results.getInt(YEAR_COLUMN));
				track.duration(
						validate(results.getString(DURATION_COLUMN)));
				track.bitrate(
						results.getInt(BITRATE_COLUMN));
				track.samplerate(
						results.getInt(SAMPLERATE_COLUMN));
				track.format(
						validate(results.getString(FORMAT_COLUMN)));
				track.comment(
						validate(results.getString(COMMENT_COLUMN)));
				disc.addTrack(track);
			}
			
			results.close();
			statement.close();
		}
		catch(SQLException e)
		{
			e.printStackTrace();
		}
		
		return disc;
	}
	
	/**
	 * Deletes a backup from the database if it exists.
	 * @param backupid - backupid of the job to delete.
	 * @return true iff the backup existed and was 
	 * successfully deleted. False otherwise.
	 */
	synchronized public static boolean deleteBackup(int backupid)
	{
		boolean success = true;
		
		try
		{
			Statement statement = conn.createStatement();
			String sql = 
				"delete from " + DBConsts.BACKUP_TABLE + 
				" where " + DBConsts.BACKUP_BACKUP_ID + " = " + backupid + ";";
			if(statement.executeUpdate(sql) != 1)
			{
				System.out.println("Nothing deleted");
				success = false;
			}
			
			try
			{
				sql = "drop table " + DBConsts.INDIVIDUAL_DISC_TABLE + backupid + ";";
				statement.execute(sql);
			}
			catch(SQLException e)
			{
				//The table was not there to delete.
				e.printStackTrace();
			}
			
			statement.close();
		}
		catch(SQLException e)
		{
			e.printStackTrace();
			success = false;
		}
		
		return success;
	}
	
	/**
	 * when a null string is put into the database, it
	 * shows up as the string "null". If we get this
	 * "null" string, converted it to a null reference
	 * to maintain the null-ness of the string.
	 * @param string - string to validate or convert
	 * to null.
	 * @return validated or nullified string.
	 */
	private static String validate(String string)
	{
		String validated = string;
		if(string != null && string.equals("null"))
		{
			validated = null;
		}
		return validated;
	}
}
