package org.gnu.composer.PrinterAccess;

import java.io.*;
import java.net.*;

public class PrinterClient 
{
	public static final String PRINTER_SERVER_ADDR = "localhost";
	public static final int PRINTER_SERVER_PORT    = 30000;
	public static final int MAX_RESPONSE_LENGTH    = 102;
	
	private Socket socket_;
	private InputStream in_;
	private OutputStream out_;
	
	private boolean serverAlive_ = false;
	
	private static final int COMMAND_INDEX    = 0;
	private static final int ARG_LENGTH_INDEX = 1;
	private static final int ARGS_START_INDEX = 2;
	
	private static final int STD_CMD_SIZE         = 10;
	private static final int STD_CMD_CKSUM_INDEX  = 9;
	private static final int STD_CMD_RESPONSE_LEN = 3;
	
	private static final int RAW_CMD_BASE_SIZE    = 3;
	
	private static final int STATUS_QUERY_SIZE    = 2;
	private static final int STATUS_RESPONSE_SIZE = 102;
	private static final int TOTAL_STATUS_BYTES   = 100;
	
	private static final int PRINT_PRN_RESPONSE_SIZE = 3;
	
	private static final int MAX_ARG_SIZE         = 255;
	
	private static final byte SUCCESS = 1;
	private static final byte FAILURE = 0;
	
	/**
	 * Example program to show how to use the PrinterClient
	 * @param args
	 */
	public static void main(String[] args) 
	{
		boolean success = true;
		PrinterClient client = PrinterClient.getSingleton();
		client.initialize();
		
		if(success)
		{
			success = client.sendStdCommand((byte)153);
		}
		
		if(success)
		{
			System.out.println("Successful in sending command to Bravopro");
		}
		else
		{
			System.out.println("Failed to send command to Bravopro");
		}
		
		int [] status = client.getStatus();
		status = client.getStatus();
		status = client.getStatus();
		status = client.getStatus();
		status = client.getStatus();
		
		if(status != null)
		{
			System.out.print("Status: ");
			for(int index = 0; index < status.length; index++)
			{
				System.out.print(status[index] + " ");
			}
		}
		else
		{
			System.out.println("Status is null");
		}
		
		if(client.printPrnFile("/home/quatre/Composer/Composer/cfg/Print Files/reject_mac.prn"))
		{
			System.out.println("Print success");
		}
		else
		{
			System.out.println("Print failure");
		}
		
		client.shutdown();
	}    
	
	private static PrinterClient singleton_;
	
	/**
	 * Load up the printer client.
	 */
	static
	{
		singleton_ = new PrinterClient();
	}
	
	/**
	 * do not let anyone make a print client object
	 */
	private PrinterClient()
	{
		
	}
	
	/**
	 * Get a hold of the singleton
	 */
	public static PrinterClient getSingleton()
	{
		return singleton_;
	}
	
	/**
	 * Initialize the Printer client. You must do this
	 * and have it successfully connect before using the 
	 * PrinterClient. Initializing more than once will
	 * also have negative effects.
	 * @return
	 */
	synchronized public boolean initialize()
	{
		boolean success = false;
		
		try
		{
			socket_      = new Socket(PRINTER_SERVER_ADDR,PRINTER_SERVER_PORT);
			in_          = socket_.getInputStream();
			out_         = socket_.getOutputStream();
			success      = true;
			serverAlive_ = true;
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
		
		return success;
	}
	
	synchronized public boolean serverAlive()
	{
		return serverAlive_;
	}
	
	/**
	 * Send a standard command to the Server.
	 * @param command - standard command number
	 * @return true iff the command was sent to
	 * the server successfully then to the
	 * printer successfully.
	 */
	synchronized public boolean sendStdCommand(byte command)
	{
		boolean success = true;
		byte sizeOfStdPrinterCommand = 8;
		byte [] msg = new byte[STD_CMD_SIZE];
		msg[COMMAND_INDEX]       = CommandTypes.SEND_COMMAND;
		msg[ARG_LENGTH_INDEX]    = sizeOfStdPrinterCommand;
		msg[2]                   = 27;
		msg[3]                   = 4;
		msg[4]                   = command;
		msg[5]                   = 0;
		msg[6]                   = 0;
		msg[7]                   = 0;
		msg[8]                   = 0;
		msg[STD_CMD_CKSUM_INDEX] = 0;
		
		int checksum = 0;
		//calculate checksum
		for(int index = ARGS_START_INDEX; index < msg.length; index++)
		{
			checksum += ByteMath.convByteToInt(msg[index]);
		}
		
		msg[STD_CMD_CKSUM_INDEX] = (byte)checksum;
		
		success = send(msg);
		
		if(success)
		{
			byte [] response = recv();
			if(response != null)
			{
				if(response.length == STD_CMD_RESPONSE_LEN)
				{
					if(response[ARGS_START_INDEX] != SUCCESS)
					{
						System.out.println("Message did not make it to the BravoPro");
						success = false;
					}
				}
				else
				{
					System.out.println("Invalid response size for std command: " + response.length);
					success = false;
				}
			}
			else
			{
				success      = false;
			}
		}
		
		return success;
	}
	
	/**
	 * Sends a raw command to the PrinterServer
	 * @param command - array of bytes to send
	 * @param calculateChecksumAsLastByte - option to
	 * calculate the checksum of all command bytes and
	 * append the checksum as one more byte at the end.
	 * @return result of operation
	 */
	synchronized public boolean sendRawCommand(byte [] command, boolean calculateChecksumAsLastByte)
	{
		boolean success = true;
		
		int argSize = command.length;
		if(calculateChecksumAsLastByte)
		{
			argSize += 1;
		}
		
		int requestSize = RAW_CMD_BASE_SIZE + argSize;
		
		byte [] msg = new byte[requestSize];
		msg[COMMAND_INDEX]       = CommandTypes.SEND_COMMAND;
		msg[ARG_LENGTH_INDEX]    = (byte)argSize;
		
		System.arraycopy(command, 0, msg, ARGS_START_INDEX, command.length);
		
		if(calculateChecksumAsLastByte)
		{
			//Determine the checksum and put it in the last index
			int checksum = ByteMath.calcChecksum(command);
			msg[msg.length - 1] = (byte)checksum;
		}
		
		success = send(msg);
		
		if(success)
		{
			byte [] response = recv();
			if(response != null)
			{
				if(response.length == STD_CMD_RESPONSE_LEN)
				{
					if(response[ARGS_START_INDEX] != SUCCESS)
					{
						System.out.println("Message did not make it to the BravoPro");
						success = false;
					}
				}
				else
				{
					System.out.println("Invalid response size for std command: " + response.length);
					success = false;
				}
			}
			else
			{
				success = false;
			}
		}
		
		return success;
	}
	
	/**
	 * Sends a raw command to the PrinterServer.
	 * The command bytes are added up, and a checksum byte
	 * is appended to the end before sending.
	 * @param command - command to send
	 * @return result of operation.
	 */
	synchronized public boolean sendRawCommand(byte [] command)
	{
		final boolean CALCULATE_CHECKSUM = true;
		return sendRawCommand(command, CALCULATE_CHECKSUM);
	}
	
	/**
	 * Get the status of the printer as an array of 100 integers.
	 * @return array of 100 integers, or null if the printer
	 * could not be queried due to:
	 * 1) printer missing
	 * 2) server not connected
	 */
	synchronized public int [] getStatus()
	{
		int [] status = null;
		boolean success;
		
		byte [] statusQuery = new byte[STATUS_QUERY_SIZE];
		statusQuery[COMMAND_INDEX] = CommandTypes.GET_STATUS;
		statusQuery[ARG_LENGTH_INDEX] = 0;
		
		success = send(statusQuery);
		
		if(success)
		{
			byte [] response = recv();
			if(response != null)
			{
				if(response.length == STATUS_RESPONSE_SIZE)
				{
					status = new int[TOTAL_STATUS_BYTES];
					int statusIndex = 0;
					for(int responseIndex = ARGS_START_INDEX; responseIndex < STATUS_RESPONSE_SIZE; responseIndex++)
					{
						status[statusIndex++] = ByteMath.convByteToInt(response[responseIndex]);
					}
				}
				else
				{
					System.out.println("Invalid response size for std command: " + response.length);
					success = false;
				}
			}
			else
			{
				success = false;
			}
		}
		
		return status;
	}
	
	/**
	 * Send request to print a prn file on a disc which
	 * is already in the printer.
	 * @param prnFile - path and file name to the prn file.
	 * @return true iff the request was successfully sent
	 * to the printer. This does not guarantee that the
	 * print occured, only that the request was correctly
	 * relayed.
	 */
	synchronized public boolean printPrnFile(String prnFile)
	{
		boolean success = true;
		
		byte [] request;
		
		byte [] prnFileNameBytes = prnFile.getBytes();
		if(prnFileNameBytes.length > MAX_ARG_SIZE)
		{
			success = false;
		}
		
		if(success)
		{
			//create request for command number, arg length, and arg
			request = new byte[2 + prnFileNameBytes.length];
			request[COMMAND_INDEX]    = CommandTypes.PRINT_PRN;
			request[ARG_LENGTH_INDEX] = (byte)prnFileNameBytes.length;
			System.arraycopy(prnFileNameBytes, 0, request, ARGS_START_INDEX, prnFileNameBytes.length);
			success = send(request);
		}
		
		if(success)
		{
			byte [] response = recv();
			if(response != null)
			{
				if(response.length == PRINT_PRN_RESPONSE_SIZE)
				{
					if(response[ARGS_START_INDEX] != SUCCESS)
					{
						success = false;
					}
				}
				else
				{
					success = false;
				}
			}
			else
			{
				success = false;
			}
		}
		
		return success;
	}
	
	/**
	 * Request that the PrinterServer shutdown.
	 * @return True if the shutdown messages was
	 * correctly sent to the server and the
	 * socket was successfully shutdown.
	 */
	synchronized public boolean shutdown()
	{
		boolean success;
		
		//Format the shutdown message
		byte [] shutdownMessage           = new byte[2];
		shutdownMessage[COMMAND_INDEX]    = CommandTypes.SHUTDOWN;
		shutdownMessage[ARG_LENGTH_INDEX] = 0;
		
		//Send the message
		success = send(shutdownMessage);
		
		//Close the socket.
		try
		{
			socket_.close();
		}
		catch(IOException e)
		{
			success = false;
		}
		
		return success;
	}
	
	/**
	 * Send a formatted message to the server.
	 * @param msg - formatted message
	 * @return True iff the message was sent
	 * correctly, false otherwise.
	 */
	private boolean send(byte [] msg)
	{
		boolean success;
		
		try
		{
			out_.write(msg);
			out_.flush();
			success = true;
		}
		catch(IOException e)
		{
			serverAlive_ = false;
			success = false;
			e.printStackTrace();
		}
		
		return success;
	}
	
	/**
	 * Receive a response from the server. This includes
	 * receiving the command, args length, and 
	 * @return
	 */
	private byte [] recv()
	{
		byte [] response = null;
		int command;
		int argLength;
		
		try
		{
			//Get the command
			command = in_.read();
			
			//get arg length
			argLength = in_.read();
			
			if(command == -1 || argLength == -1)
			{
				System.out.println("Server connection ended");
				serverAlive_ = false;
			}
			else
			{
				//Create the response array with room for
				//the command, arg length, and anticipated
				//argument
				response                   = new byte[argLength + 2];
				response[COMMAND_INDEX]    = (byte)command;
				response[ARG_LENGTH_INDEX] = (byte)argLength;
				
				//If this response has any arguments, gather them up
				if(argLength > 0)
				{
					int readBytes = 
						in_.read(response, ARGS_START_INDEX, argLength);
					
					if(readBytes != argLength)
					{
						System.out.println(
							"Needed arg of " + argLength + " bytes but got "
							+ readBytes + " bytes");
						response = null;
						serverAlive_ = false;
					}
				}
			}
		}
		catch(IOException e)
		{
			response     = null;
			serverAlive_ = false;
			e.printStackTrace();
		}
		
		return response;
	}
}

/**
 * Do not modify these values without also modifying the
 * values in PrinterTools.h. These values are mirrored
 * from that file.
 * 
 * @author Branden Archer
 */
class CommandTypes
{
	   public static final int SEND_COMMAND = 1;
	   public static final int GET_STATUS   = 2;
	   public static final int PRINT_PRN    = 3;
       public static final int PRINT_FILE   = 4;
	   public static final int SHUTDOWN     = 5;
}
