
#ifndef PRINTERSERVER_H
#define PRINTERSERVER_H

//Do not modify these values without also modifying the
//values in PrinterClient.java. These values are mirrored
//from that file.
enum CommandTypes
{
   SEND_COMMAND = 1,
   GET_STATUS   = 2,
   PRINT_PRN    = 3,
   PRINT_IMAGE  = 4,
   SHUTDOWN     = 5,
};

#define SUCCESS 1
#define FAILURE 0

#define SERVER_PORT 30000
#define SERVER_ADDR "127.0.0.1"

#define COMMAND_HEADER_SIZE 1
#define ARG_SIZE_HEADER_SIZE 1
#define MAX_MSG_SIZE 255
#define MAX_CONNECTS_TO_QUEUE 1

#define TOTAL_STATUS_BYTES 100

/**
 * Binds to the client for the session.
 * This includes all setup to the OS for
 * the socket, then the binding to the
 * port, and finally accepting the
 * connection from the client. This method
 * will block until all of this is completed.
 */
void bindToClient();

/**
 * Binds to the printer at the path given at
 * program start. This method will not exit
 * until the printer has successfully been
 * opened up for communication.
 */
void bindToPrinter();

/**
 * This is the main server loop. This method
 * will continue to accept requests from the
 * user. If the user sends a request and the
 * printer does not respond, this information
 * is sent back to the user and this function
 * returns. If the user requests that this
 * server shutdown this function will terminate.
 */
void serveRequests();

/**
 * Receives a command from the other. This
 * command has three components: 
 *
 * 1) Command (1 byte)
 * 2) Size of argument (1 byte)
 * 3) Argument (depends on #2, 0-255 bytes)
 *
 * When a command is received, the three components
 * are put into the global variables:
 * 1) commandHeader
 * 2) argumentSizeHeader
 * 3) data
 *
 * When this function returns, either all of the
 * message has been stored into the global variables,
 * or the client has terminated the connection.
 * 
 * @return True iff the data was correctly received
 * from the client. False if a communication error
 * occured or the client terminated the connection.
 */
bool recvFromClient();

/**
 * Sends a response to the client. This response
 * has three components:
 *
 * 1) Command (1 byte)
 * 2) Size of argument (1 byte)
 * 3) Argument (depends on #2, 0-255 bytes)
 *
 * The command is derived from three global
 * variables:
 * 1) commandHeader
 * 2) argumentSizeHeader
 * 3) data
 *
 * @return True iff the send is successful,
 * False otherwise.
 */
bool sendResponse();

/**
 * Read a number of bytes from the printer. These
 * bytes will represent the status of the printer.
 * If the printer is not connected, this function
 * will return null. Otherwise it is assumed the
 * status was retrieved correctly.
 * @return True iff the status was written to the
 * global memory space, false if the printer is
 * not attached.
 */
bool readStatus();

/**
 * Sends a command to the printer. There is no
 * indication if this send was successful or not,
 * so it is recommended to do a status gather after
 * sending the command, as the status will fail if the
 * printer is not attached.
 * @param command - array of bytes to send to the printer.
 * @param sizeOfCommand - size of command in bytes
 */
void sendCommand(unsigned char * command, int sizeOfCommand);

/**
 * Prints a prn file to the printer.
 * @return True if the operation was a
 * success, and the printer received the
 * data.
 */
bool sendPrnFile(unsigned char * prnFileName);

#endif
